import os

import torch
import torch.nn as nn

from diffusers.models.attention_processor import (
    Attention,
    AttnProcessor,
    AttnProcessor2_0,
)

from .attention_processor import FusedSelfAttnProcessor

_IS_ONEDIFF_QUANT_AVAILABLE = 0
try:
    import onediff_quant
    from onediff_quant import DynamicQuantLinearModule, StaticQuantLinearModule

    _IS_ONEDIFF_QUANT_AVAILABLE = 1
except ImportError as e:
    pass


def rewrite_self_attention(model):
    if "GeForce RTX" in torch.cuda.get_device_name():
        return

    def is_attention_can_be_fused(attn):
        dim_head = attn.to_q.out_features // attn.heads
        if dim_head not in (40, 64):
            return False
        if attn.to_k is None or attn.to_v is None:
            return False
        if (
            attn.to_q.bias is not None
            or attn.to_k.bias is not None
            or attn.to_v.bias is not None
        ):
            return False
        if not isinstance(attn.processor, (AttnProcessor, AttnProcessor2_0)):
            return False
        if (
            attn.to_q.in_features != attn.to_k.in_features
            or attn.to_q.in_features != attn.to_v.in_features
        ):
            return False
        if not (
            attn.to_q.weight.dtype == attn.to_k.weight.dtype
            and attn.to_q.weight.dtype == attn.to_v.weight.dtype
        ):
            return False
        return True

    def rewrite_attention(attn):
        dim_head = attn.to_q.out_features // attn.heads
        has_bias = attn.to_q.bias is not None
        attn.to_qkv = nn.Linear(
            attn.to_q.in_features, attn.to_q.out_features * 3, bias=has_bias
        )
        attn.to_qkv.requires_grad_(False)

        qkv_weight = torch.cat(
            [
                attn.to_q.weight.permute(1, 0).reshape(-1, attn.heads, dim_head),
                attn.to_k.weight.permute(1, 0).reshape(-1, attn.heads, dim_head),
                attn.to_v.weight.permute(1, 0).reshape(-1, attn.heads, dim_head),
            ],
            dim=2,
        )
        qkv_weight = (
            qkv_weight.reshape(-1, attn.to_q.out_features * 3)
            .permute(1, 0)
            .contiguous()
        )
        attn.to_qkv.weight.data = qkv_weight

        if has_bias:
            qkv_bias = (
                torch.cat(
                    [
                        attn.to_q.bias.reshape(attn.heads, dim_head),
                        attn.to_k.bias.reshape(attn.heads, dim_head),
                        attn.to_v.bias.reshape(attn.heads, dim_head),
                    ],
                    dim=1,
                )
                .reshape(attn.to_q.out_features * 3)
                .contiguous()
            )
            attn.to_qkv.bias.data = qkv_bias

        if _IS_ONEDIFF_QUANT_AVAILABLE and isinstance(
            attn.to_q, (DynamicQuantLinearModule, StaticQuantLinearModule)
        ):
            cls = type(attn.to_q)
            weight_scale = (
                torch.cat(
                    [
                        torch.Tensor(attn.to_q.calibrate[2]).reshape(
                            attn.heads, dim_head
                        ),
                        torch.Tensor(attn.to_k.calibrate[2]).reshape(
                            attn.heads, dim_head
                        ),
                        torch.Tensor(attn.to_v.calibrate[2]).reshape(
                            attn.heads, dim_head
                        ),
                    ],
                    dim=1,
                )
                .reshape(attn.to_q.out_features * 3)
                .contiguous()
            )
            calibrate = [attn.to_q.calibrate[0], attn.to_q.calibrate[1], weight_scale]
            attn.to_qkv = cls(attn.to_qkv, attn.to_q.nbits, calibrate, attn.to_q.name)

        del attn.to_q
        del attn.to_k
        del attn.to_v
        attn.processor = FusedSelfAttnProcessor(attn)

    for m in model.modules():
        if isinstance(m, Attention) and is_attention_can_be_fused(m):
            rewrite_attention(m)



import importlib
import os
import sys
from functools import lru_cache
from importlib.metadata import requires
from inspect import ismodule
from pathlib import Path
from types import FunctionType, ModuleType
from typing import Optional, Union

from onediff.utils import logger
from .dyn_mock_mod import DynamicMockModule
from .format_utils import MockEntityNameFormatter

__all__ = ["LazyMocker", "is_need_mock"]

# Cache all imported modules (maxsize=None: no limit)
@lru_cache(maxsize=None)
def has_torch_dependency(main_pkg: str):
    try:
        pkgs = requires(main_pkg)
    except Exception as e:
        # packages may lack metadata
        return False
    return any(pkg.split(" ")[0] == "torch" for pkg in pkgs)


def is_need_mock(cls) -> bool:
    assert isinstance(cls, (type, str))
    main_pkg = cls.__module__.split(".")[0]

    if main_pkg == "torch":
        return True

    return has_torch_dependency(main_pkg)


class DynamicModuleLoader(ModuleType):
    def __init__(self, obj_entity: ModuleType, pkg_root=None, module_path=None):
        self.obj_entity = obj_entity
        self.pkg_root = pkg_root
        self.module_path = module_path

    @classmethod
    def from_path(cls, module_path: str):
        model_name = os.path.basename(module_path)
        module = importlib.import_module(model_name)
        return cls(module, module_path, module_path)

    def __getattr__(self, name):
        obj_entity = getattr(self.obj_entity, name, None)
        module_path = os.path.join(self.module_path, name)
        if obj_entity is None:
            pkg_name = os.path.basename(self.pkg_root)
            absolute_name = os.path.relpath(module_path, self.pkg_root).replace(
                os.path.sep, "."
            )
            absolute_name = f"{pkg_name}.{absolute_name}"
            obj_entity = importlib.import_module(absolute_name)
        if ismodule(obj_entity):
            return DynamicModuleLoader(obj_entity, self.pkg_root, module_path)
        return obj_entity


class LazyMocker:
    def __init__(self, prefix: str, suffix: str, tmp_dir: Optional[Union[str, Path]]):
        self.prefix = prefix
        self.suffix = suffix
        self.tmp_dir = tmp_dir
        self.mocked_packages = set()
        self.cleanup_list = []

    def mock_package(self, package: str):
        pass

    def cleanup(self):
        pass

    def get_mock_entity_name(self, entity: Union[str, type, FunctionType]):
        formatter = MockEntityNameFormatter(prefix=self.prefix, suffix=self.suffix)
        full_obj_name = formatter.format(entity)
        return full_obj_name

    def mock_entity(self, entity: Union[str, type, FunctionType]):
        """Mock the entity and return the mocked entity

        Example:
            >>> mocker = LazyMocker(prefix="mock_", suffix="_of", tmp_dir="tmp")
            >>> mocker.mock_entity("models.DemoModel")
            <class 'mock_models_of.DemoModel'>
            >>> cls_obj = models.DemoModel
            >>> mocker.mock_entity(cls_obj)
            <class 'mock_models_of.DemoModel'>
        """
        return self.load_entity_with_mock(entity)

    def add_mocked_package(self, package: str):
        if package in self.mocked_packages:
            return

        self.mocked_packages.add(package)
        package = sys.modules.get(package, None)

        # TODO remove code below
        # fix the mock error in https://github.com/siliconflow/oneflow/blob/main/python/oneflow/mock_torch/mock_importer.py#L105-L118
        if package and getattr(package, "__file__", None) is not None:
            pkg_path = Path(package.__file__).parents[1]
            if pkg_path not in sys.path:
                sys.path.append(str(pkg_path))

    def load_entity_with_mock(self, entity: Union[str, type, FunctionType]):
        formatter = MockEntityNameFormatter(prefix=self.prefix, suffix=self.suffix)
        full_obj_name = formatter.format(entity)
        attrs = full_obj_name.split(".")
        if attrs[0] == "__main__":
            import __main__

            main_path = __main__.__file__
            main_path_parent = Path(main_path).parent
            if str(main_path_parent) not in sys.path:
                sys.path.append(str(main_path_parent))
            main_name = Path(main_path).stem
            mock_main = DynamicMockModule.from_package(main_name, verbose=False)
            for name in attrs[1:]:
                mock_main = getattr(mock_main, name)
            return mock_main

        # add package path to sys.path to avoid mock error
        self.add_mocked_package(attrs[0])

        mock_pkg = DynamicMockModule.from_package(attrs[0], verbose=False)
        for name in attrs[1:]:
            mock_pkg = getattr(mock_pkg, name)
        return mock_pkg



import importlib
import inspect
import os
from contextlib import contextmanager
from copy import deepcopy
from inspect import ismodule, signature
from types import ModuleType
from typing import Dict, List

import torch
from oneflow.mock_torch import enable
from oneflow.mock_torch.mock_importer import _importer

from onediff.utils import logger
from .import_module_utils import import_module_from_path
from .patch_for_compiler import *

__all__ = ["DynamicMockModule"]


def _get_module(full_name: str):
    try:
        attrs = full_name.split(".")
        module = importlib.import_module(attrs[0])
        for attr in attrs[1:]:
            module = getattr(module, attr)
        return module
    except Exception as e:
        pass


def inspect_modules_and_attributes(module_names):
    all_results = {}

    def get_attribute_info(module_name):
        module = _get_module(module_name)
        if module is None:
            return {}
        attribute_info = {}
        for attr in (a for a in dir(module) if not a.startswith("__")):
            try:
                attr_value = getattr(module, attr)
                source_file = inspect.getsourcefile(attr_value)
                attribute_info[f"{module_name};{attr}"] = (source_file, f"{attr_value}")
            except Exception as e:
                pass

        return attribute_info

    for name in module_names:
        all_results.update(get_attribute_info(name))

    return all_results


def getattr_from_module_name(module, module_name: str):
    # case <function get_additional_models_factory.<locals>.get_additional_models_with_motion at 0x7fd0542cf6d0>
    full_attr_name = module_name.split(" ")[1]
    attrs = full_attr_name.split(".")
    sub_module = module
    for attr in attrs:
        if attr == "<locals>":
            if len(signature(sub_module).parameters) == 0:
                logger.info(f"{full_attr_name=} is a local function without parameters")

                def proxy_func(*args, **kwargs):
                    return sub_module()(*args, **kwargs)

                return proxy_func
            else:
                # logger.warning(
                #     f"Not support {module_name} with parameters Module: {module}"
                # )
                raise RuntimeError(
                    f"Not support {module_name} with parameters Module: {module}"
                )
        sub_module = getattr(sub_module, attr, None)
    return sub_module


def _update_module(full_names, main_pkg_enable_context):
    with main_pkg_enable_context():
        original_results = inspect_modules_and_attributes(full_names)

    updated_results = inspect_modules_and_attributes(full_names)

    torch_path = os.path.dirname(torch.__file__)

    for module_key, (module_path, module_code) in updated_results.items():
        org_module_path, org_module_code = original_results.get(
            module_key, (module_path, module_code)
        )

        if org_module_path != module_path or org_module_code != module_code:
            # Skip torch module Because torch module is already mocked by oneflow
            if torch_path in module_path:
                continue

            # Update module inplace
            module_name, attr_name = module_key.split(";")
            # sample_module = importlib.import_module(module_name).__dict__[attr_name]
            good_module = _get_module(module_name)
            sample_module = getattr(good_module, attr_name)
            package_space = inspect.getmodule(sample_module).__name__
            with main_pkg_enable_context():
                module = _get_module(module_name)
                if package_space == "__main__":
                    other = import_module_from_path(module_path)
                else:
                    other = importlib.import_module(package_space)

                value = getattr_from_module_name(other, module_name=str(sample_module))
                if value is None:
                    continue

                setattr(module, attr_name, value)


class DynamicMockModule(ModuleType):
    def __init__(
        self,
        pkg_name: str,
        obj_entity: ModuleType,
        main_pkg_enable: callable,
    ):
        self._pkg_name = pkg_name
        self._obj_entity = obj_entity  # ModuleType or _LazyModule
        self._main_pkg_enable = main_pkg_enable
        self._intercept_dict = {}

    def __repr__(self) -> str:
        return f"<DynamicMockModule {self._pkg_name} {self._obj_entity}>"

    def hijack(self, module_name: str, obj: object):
        self._intercept_dict[module_name] = obj

    @classmethod
    def from_package(
        cls,
        main_pkg: str,
        *,
        lazy: bool = True,
        verbose: bool = False,
        extra_dict: Dict[str, str] = None,
        required_dependencies: List[str] = [],
    ):
        assert isinstance(main_pkg, str)

        @contextmanager
        def main_pkg_enable():
            with enable(
                lazy=lazy,
                verbose=verbose,
                extra_dict=extra_dict,
                main_pkg=main_pkg,
                mock_version=True,
                required_dependencies=required_dependencies,
            ):
                yield

        with main_pkg_enable():
            obj_entity = importlib.import_module(main_pkg)
        return cls(main_pkg, obj_entity, main_pkg_enable)

    def _get_module(self, _name: str):
        # Fix Lazy import
        # https://github.com/huggingface/diffusers/blob/main/src/diffusers/__init__.py#L728-L734
        module_name = f"{self._obj_entity.__name__}.{_name}"
        try:
            return importlib.import_module(module_name)
        except Exception as e:
            raise RuntimeError(
                f"Failed to import {module_name} because of the following error (look up to see its"
                f" traceback):\n{e}"
            ) from e

    def __getattr__(self, name: str):
        fullname = f"{self._obj_entity.__name__}.{name}"
        if fullname in self._intercept_dict:
            return self._intercept_dict[fullname]

        with self._main_pkg_enable():
            obj_entity = getattr(self._obj_entity, name, None)
            if obj_entity is None:
                obj_entity = self._get_module(name)
            org_delete_list = deepcopy(_importer.delete_list)

        try:
            # Update obj_entity inplace
            if not _importer.enable:
                _update_module([fullname] + org_delete_list, self._main_pkg_enable)
        except Exception as e:
            logger.debug(f"Failed to update obj_entity in place. Exception: {e}")

        if ismodule(obj_entity):
            return DynamicMockModule(self._pkg_name, obj_entity, self._main_pkg_enable)

        return obj_entity

    def __all__(self):
        with self._main_pkg_enable():
            return dir(self._obj_entity)



import importlib
import logging
import os
import types
import warnings
from pathlib import Path
from typing import Dict, List, Union

from onediff.utils import logger
from ..import_tools.importer import LazyMocker

__all__ = ["transform_mgr"]


class TransformManager:
    """TransformManager

    __init__ args:
        `debug_mode`: Whether to print debug info.
        `tmp_dir`: The temp dir to store mock files.
    """

    def __init__(self, debug_mode=False, tmp_dir="./output"):
        self.debug_mode = debug_mode
        self._torch_to_oflow_cls_map = {}
        self._oflow_to_torch_cls_map = {}
        self._setup_logger()
        self.mocker = LazyMocker(prefix="", suffix="", tmp_dir=None)
        self.loaded_modules = set()

    def _setup_logger(self):
        name = "ONEDIFF"
        level = logging.DEBUG if self.debug_mode else logging.WARNING
        logger.configure_logging(name=name, file_name=None, level=level, log_dir=None)
        self.logger = logger

    def get_mocked_packages(self):
        return self.mocker.mocked_packages

    def load_class_proxies_from_packages(self, package_names: List[Union[Path, str]]):
        self.logger.debug(f"Loading modules: {package_names}")
        for package_name in package_names:
            self.mocker.mock_package(package_name)
            self.logger.info(f"Loaded Mock Torch Package: {package_name} successfully")

    def update_class_proxies(self, class_proxy_dict: Dict[str, type], verbose=True):
        """Update `_torch_to_oflow_cls_map` with `class_proxy_dict`.

        example:
            `class_proxy_dict = {"mock_torch.nn.Conv2d": flow.nn.Conv2d}`

        """
        self._torch_to_oflow_cls_map.update(class_proxy_dict)

        debug_message = f"Updated class proxies: {len(class_proxy_dict)=}"
        debug_message += f"\n{class_proxy_dict}\n"
        self.logger.debug(debug_message)

    def _transform_entity(self, entity):
        # TODO: Optimize _transform_entity for faster SDXL conversion (1.47s)
        result = self.mocker.mock_entity(entity)
        if result is None:
            RuntimeError(f"Failed to transform entity: {entity}")
        return result

    def get_transformed_entity_name(self, entity):
        return self.mocker.get_mock_entity_name(entity)

    def transform_cls(self, cls):
        """Transform a class to a mock class ."""
        full_cls_name = cls.__module__ + "." + cls.__qualname__
        mock_full_cls_name = self.get_transformed_entity_name(full_cls_name)

        # transform cache
        if mock_full_cls_name in self._torch_to_oflow_cls_map:
            return self._torch_to_oflow_cls_map[mock_full_cls_name]

        # transform
        if cls.__module__.startswith("torch."):
            mod_name = cls.__module__.replace("torch.", "oneflow.")
            mod = importlib.import_module(mod_name)
            mock_cls = getattr(mod, cls.__name__)
        else:
            mock_cls = self._transform_entity(mock_full_cls_name)

        self._torch_to_oflow_cls_map[mock_full_cls_name] = mock_cls
        self._oflow_to_torch_cls_map[mock_full_cls_name] = cls
        return mock_cls

    def reverse_transform_cls(self, cls):
        full_cls_name = cls.__module__ + "." + cls.__qualname__
        mock_full_cls_name = self.get_transformed_entity_name(full_cls_name)
        if mock_full_cls_name in self._oflow_to_torch_cls_map:
            return self._oflow_to_torch_cls_map[mock_full_cls_name]
        else:
            self.logger.info(
                f"{mock_full_cls_name} not in _oflow_to_torch_cls_map, import it directly."
            )
            mod = importlib.import_module(cls.__module__)
            return getattr(mod, cls.__qualname__)

    def transform_func(self, func: types.FunctionType):
        mock_full_cls_name = self.get_transformed_entity_name(func)
        if mock_full_cls_name in self._torch_to_oflow_cls_map:
            return self._torch_to_oflow_cls_map[mock_full_cls_name]
        return self._transform_entity(func)

    def transform_package(self, package_name):
        return self._transform_entity(package_name)


debug_mode = os.getenv("ONEDIFF_DEBUG", "0") == "1"
transform_mgr = TransformManager(debug_mode=debug_mode, tmp_dir=None)

if not transform_mgr.debug_mode:
    warnings.simplefilter("ignore", category=UserWarning)
    warnings.simplefilter("ignore", category=FutureWarning)


if importlib.util.find_spec("pydantic") is not None:
    import pydantic

    if pydantic.VERSION < "2.5.2":
        logger.warning(
            f"Pydantic version {pydantic.VERSION} is too low, please upgrade to 2.5.2 or higher."
        )
        from oneflow.mock_torch.mock_utils import MockEnableDisableMixin

        MockEnableDisableMixin.hazard_list.append(
            "huggingface_hub.inference._text_generation"
        )



from onediff.infer_compiler.backends.oneflow.transform import register

import oneflow as flow  # usort: skip
import diffusers_enterprise_lite

torch2oflow_class_map = {
    diffusers_enterprise_lite.ProxyFakeModule: diffusers_enterprise_lite.OneFlowFakeHPCModule,
    diffusers_enterprise_lite.ProxyDynamicConvModule: diffusers_enterprise_lite.OneFlowDynamicHPCConvModule,
    diffusers_enterprise_lite.ProxyDynamicLinearModule: diffusers_enterprise_lite.OneFlowDynamicHPCLinearModule,
}


def convert_func(mod: flow.Tensor, verbose=False):
    return mod


register(torch2oflow_class_map=torch2oflow_class_map, torch2oflow_funcs=[convert_func])



import math

import torch
import oneflow as flow  # usort: skip
import oneflow.nn.functional as F


class FakeCuda:
    @staticmethod
    def current_device():
        return "cuda:0"

    @staticmethod
    def mem_get_info(dev):
        return torch.cuda.mem_get_info(str(dev))

    @staticmethod
    def _scaled_dot_product_attention_math(
        query, key, value, attn_mask=None, dropout_p=0.0, is_causal=False
    ):
        d_k = query.size(-1)

        if is_causal:
            assert attn_mask is None, "Cannot use both attn_mask and is_causal=True"
            L, S = query.size(-2), key.size(-2)
            attn_mask = flow.ones((L, S), dtype=flow.bool).tril()

        if attn_mask is not None:
            if attn_mask.dtype == flow.bool:
                new_attn_mask = flow.empty(
                    attn_mask.shape, dtype=query.dtype, device=query.device
                )
                mask = flow.logical_not(attn_mask)
                new_attn_mask.masked_fill_(mask, float("-inf"))
                attn_mask = new_attn_mask

        scores = flow.matmul(query, key.transpose(-2, -1)) / math.sqrt(d_k)

        if attn_mask is not None:
            scores.add_(attn_mask)

        p_attn = F.softmax(scores, dim=-1)

        if dropout_p > 0.0:
            generator = flow.Generator()
            p_attn = flow.nn.functional.dropout(
                p_attn, p=dropout_p, generator=generator
            )

        return flow.matmul(p_attn, value)

    @staticmethod
    def scaled_dot_product_attention(
        query, key, value, attn_mask=None, dropout_p=0.0, is_causal=False
    ):
        """Scaled Dot-Product Attention
        Args:
        query (Tensor): Query tensor; shape :math:`(N, ..., L, E)`.
        key (Tensor): Key tensor; shape :math:`(N, ..., S, E)`.
        value (Tensor): Value tensor; shape :math:`(N, ..., S, Ev)`.
        attn_mask (optional Tensor): Attention mask; shape :math:`(N, ..., L, S)`. Two types of masks are supported.
            A boolean mask where a value of True indicates that the element *should* take part in attention.
            A float mask of the same type as query, key, value that is added to the attention score.
        dropout_p (float): Dropout probability; if greater than 0.0, dropout is applied
        is_causal (bool): If true, assumes causal attention masking and errors if both attn_mask and is_causal
            are set.

        Returns:
            output (Tensor): Attention output; shape :math:`(N, ..., L, Ev)`.

        Shape legend:
            - :math:`N: \text{Batch size} ... : \text{Any number of other batch dimensions (optional)}`
            - :math:`S: \text{Source sequence length}`
            - :math:`L: \text{Target sequence length}`
            - :math:`E: \text{Embedding dimension of the query and key}`
            - :math:`Ev: \text{Embedding dimension of the value}`
        """
        if attn_mask is not None or dropout_p > 0.0:
            return FakeCuda._scaled_dot_product_attention_math(
                query, key, value, attn_mask, dropout_p, is_causal
            )

        batch_size, num_heads, target_seq_len, head_dim = query.shape
        out = flow._C.fused_multi_head_attention_inference_v2(
            query=query,
            query_layout="BHMK",
            query_head_size=head_dim,
            key=key,
            key_layout="BHMK",
            value=value,
            value_layout="BHMK",
            output_layout="BM(HK)",
            causal=is_causal,
        )
        # (N, L, H x Ev) -> (N, H, L, Ev)
        value_embed_dim = value.shape[-1]
        out = out.view(batch_size, target_seq_len, num_heads, value_embed_dim).permute(
            0, 2, 1, 3
        )
        return out


flow.cuda.current_device = FakeCuda.current_device
flow.cuda.mem_get_info = FakeCuda.mem_get_info
flow.nn.functional.scaled_dot_product_attention = FakeCuda.scaled_dot_product_attention
F.scaled_dot_product_attention = FakeCuda.scaled_dot_product_attention

flow.cuda.memory_stats = torch.cuda.memory_stats
flow.version = torch.version

from oneflow import Tensor


def oneflow_rfloordiv():
    original_rfloordiv = Tensor.__rfloordiv__

    def rfloordiv(self, other):
        if isinstance(other, int):
            other = flow.tensor(other)

        return original_rfloordiv(self, other)

    return rfloordiv


Tensor.__rfloordiv__ = oneflow_rfloordiv()



import time
from copy import deepcopy

import torch
import torch.nn as nn

from onediff.infer_compiler.backends.oneflow.transform.manager import transform_mgr
from onediff.infer_compiler.backends.oneflow.utils.cost_util import cost_cnt
from onediff.infer_compiler.backends.oneflow.utils.version_util import (
    is_quantization_enabled,
)
from onediff.torch_utils.module_operations import modify_sub_module
from onediff.utils import logger


__all__ = ["quantize_model", "varify_can_use_quantization"]


def varify_can_use_quantization():
    if not is_quantization_enabled():
        logger.warning(f"OneDiff Quantization can't be used.")
        return False
    return True


@cost_cnt(debug=transform_mgr.debug_mode)
def quantize_model(
    model,  # diffusion_model
    quantize_conv=True,
    quantize_linear=True,
    bits=8,
    *,
    inplace=True,
    calibrate_info: dict = None,
):
    """Quantize a model. inplace=True will modify the model in-place."""
    start_time = time.time()
    if varify_can_use_quantization() is False:
        return model

    from onediff_quant import Quantizer
    from onediff_quant.utils import (
        find_quantizable_modules,
        get_quantize_module,
        symm_quantize_sub_module,
    )

    quantize_conv_cnt, quantize_linear_cnt = 0, 0

    if not inplace:
        model = deepcopy(model)

    def no_quantizable(sub_module_name):
        if calibrate_info is not None:
            conf = calibrate_info.get(sub_module_name, None)
            if conf is None:
                return True
            return False
        else:
            return False

    def apply_quantization_to_modules(quantizable_modules):
        nonlocal model, quantize_conv_cnt, quantize_linear_cnt

        for sub_module_name, sub_mod in quantizable_modules.items():
            if no_quantizable(sub_module_name):
                continue

            if isinstance(sub_mod, nn.Conv2d):
                quantize_conv_cnt += 1
            elif isinstance(sub_mod, nn.Linear):
                quantize_linear_cnt += 1

            quantizer = Quantizer()
            quantizer.configure(bits=bits, perchannel=True)
            quantizer.find_params(sub_mod.weight.float(), weight=True)
            shape = [-1] + [1] * (len(sub_mod.weight.shape) - 1)
            scale = quantizer.scale.reshape(*shape)
            symm_quantize_sub_module(
                model, sub_module_name, scale, quantizer.maxq, save_as_float=False
            )

            input_scale = 0
            input_zero_point = 0
            weight_scale = scale.reshape(-1).tolist()
            sub_calibrate_info = [input_scale, input_zero_point, weight_scale]

            sub_mod = get_quantize_module(
                sub_mod,
                sub_module_name,
                sub_calibrate_info,
                fake_quant=False,
                static=False,
                nbits=bits,
            )

            modify_sub_module(model, sub_module_name, sub_mod)

    if quantize_conv:
        conv_modules = find_quantizable_modules(model, module_cls=[nn.Conv2d])
        apply_quantization_to_modules(conv_modules)
        logger.debug(f"{len(conv_modules)=}")

    if quantize_linear:
        linear_modules = find_quantizable_modules(model, module_cls=[nn.Linear])
        apply_quantization_to_modules(linear_modules)
        logger.debug(f"{len(linear_modules)=}")

    logger.info(
        f"Quantized model {type(model)} successfully! \n"
        + f"Quantized conv: {quantize_conv_cnt} \n"
        + f"Quantized linear: {quantize_linear_cnt} \n"
        + f"Time: {time.time() - start_time:.4f}s \n"
    )

    return model



import hashlib

from oneflow.framework.args_tree import ArgsTree


def extract_node_name(v):
    return type(v).__name__


def generate_input_structure_key(args_tree: ArgsTree):
    out_str = "_".join(
        (extract_node_name(node) for node in args_tree.iter_nodes() if node is not None)
    )
    return hashlib.sha256(out_str.encode("utf-8")).hexdigest()[:6]


def generate_model_structure_key(deployable_module):
    model = deployable_module._deployable_module_model.oneflow_module
    model_hash = hashlib.sha256(f"{model}".encode("utf-8")).hexdigest()
    return model_hash[:8]



import collections


class LRUCache(collections.OrderedDict):
    __slots__ = ["LEN"]

    def __init__(self, capacity: int = 9):
        self.LEN = capacity

    def get(self, key: str, default=None) -> any:
        if key in self:
            self.move_to_end(key)
            return self[key]
        else:
            return default

    def put(self, key: str, value: any) -> None:
        self[key] = value
        self.move_to_end(key)
        if len(self) > self.LEN:
            self.popitem(last=False)



import importlib.metadata
from typing import Any, Dict, List, Optional, Tuple, Union

import oneflow as torch
from onediff.infer_compiler.backends.oneflow.transform import transform_mgr
from packaging import version

diffusers_0210_v = version.parse("0.21.0")
diffusers_version = version.parse(importlib.metadata.version("diffusers"))

transformed_diffusers = transform_mgr.transform_package("diffusers")
UNet2DConditionOutput = (
    transformed_diffusers.models.unet_2d_condition.UNet2DConditionOutput
)

try:
    USE_PEFT_BACKEND = transformed_diffusers.utils.USE_PEFT_BACKEND
    scale_lora_layers = transformed_diffusers.utils.scale_lora_layers
    unscale_lora_layers = transformed_diffusers.utils.unscale_lora_layers
except Exception as e:
    USE_PEFT_BACKEND = False


class UNet2DConditionModel(
    transformed_diffusers.models.unet_2d_condition.UNet2DConditionModel
):
    def forward(
        self,
        sample: torch.FloatTensor,
        timestep: Union[torch.Tensor, float, int],
        encoder_hidden_states: torch.Tensor,
        class_labels: Optional[torch.Tensor] = None,
        timestep_cond: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        added_cond_kwargs: Optional[Dict[str, torch.Tensor]] = None,
        down_block_additional_residuals: Optional[Tuple[torch.Tensor]] = None,
        mid_block_additional_residual: Optional[torch.Tensor] = None,
        down_intrablock_additional_residuals: Optional[Tuple[torch.Tensor]] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        return_dict: bool = True,
    ) -> Union[UNet2DConditionOutput, Tuple]:
        r"""
        The [`UNet2DConditionModel`] forward method.

        Args:
            sample (`torch.FloatTensor`):
                The noisy input tensor with the following shape `(batch, channel, height, width)`.
            timestep (`torch.FloatTensor` or `float` or `int`): The number of timesteps to denoise an input.
            encoder_hidden_states (`torch.FloatTensor`):
                The encoder hidden states with shape `(batch, sequence_length, feature_dim)`.
            class_labels (`torch.Tensor`, *optional*, defaults to `None`):
                Optional class labels for conditioning. Their embeddings will be summed with the timestep embeddings.
            timestep_cond: (`torch.Tensor`, *optional*, defaults to `None`):
                Conditional embeddings for timestep. If provided, the embeddings will be summed with the samples passed
                through the `self.time_embedding` layer to obtain the timestep embeddings.
            attention_mask (`torch.Tensor`, *optional*, defaults to `None`):
                An attention mask of shape `(batch, key_tokens)` is applied to `encoder_hidden_states`. If `1` the mask
                is kept, otherwise if `0` it is discarded. Mask will be converted into a bias, which adds large
                negative values to the attention scores corresponding to "discard" tokens.
            cross_attention_kwargs (`dict`, *optional*):
                A kwargs dictionary that if specified is passed along to the `AttentionProcessor` as defined under
                `self.processor` in
                [diffusers.models.attention_processor](https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/attention_processor.py).
            added_cond_kwargs: (`dict`, *optional*):
                A kwargs dictionary containing additional embeddings that if specified are added to the embeddings that
                are passed along to the UNet blocks.
            down_block_additional_residuals: (`tuple` of `torch.Tensor`, *optional*):
                A tuple of tensors that if specified are added to the residuals of down unet blocks.
            mid_block_additional_residual: (`torch.Tensor`, *optional*):
                A tensor that if specified is added to the residual of the middle unet block.
            encoder_attention_mask (`torch.Tensor`):
                A cross-attention mask of shape `(batch, sequence_length)` is applied to `encoder_hidden_states`. If
                `True` the mask is kept, otherwise if `False` it is discarded. Mask will be converted into a bias,
                which adds large negative values to the attention scores corresponding to "discard" tokens.
            return_dict (`bool`, *optional*, defaults to `True`):
                Whether or not to return a [`~models.unet_2d_condition.UNet2DConditionOutput`] instead of a plain
                tuple.
            cross_attention_kwargs (`dict`, *optional*):
                A kwargs dictionary that if specified is passed along to the [`AttnProcessor`].
            added_cond_kwargs: (`dict`, *optional*):
                A kwargs dictionary containin additional embeddings that if specified are added to the embeddings that
                are passed along to the UNet blocks.
            down_block_additional_residuals (`tuple` of `torch.Tensor`, *optional*):
                additional residuals to be added to UNet long skip connections from down blocks to up blocks for
                example from ControlNet side model(s)
            mid_block_additional_residual (`torch.Tensor`, *optional*):
                additional residual to be added to UNet mid block output, for example from ControlNet side model
            down_intrablock_additional_residuals (`tuple` of `torch.Tensor`, *optional*):
                additional residuals to be added within UNet down blocks, for example from T2I-Adapter side model(s)

        Returns:
            [`~models.unet_2d_condition.UNet2DConditionOutput`] or `tuple`:
                If `return_dict` is True, an [`~models.unet_2d_condition.UNet2DConditionOutput`] is returned, otherwise
                a `tuple` is returned where the first element is the sample tensor.
        """
        # By default samples have to be AT least a multiple of the overall upsampling factor.
        # The overall upsampling factor is equal to 2 ** (# num of upsampling layers).
        # However, the upsampling interpolation output size can be forced to fit any upsampling size
        # on the fly if necessary.
        default_overall_up_factor = 2**self.num_upsamplers

        # upsample size should be forwarded when sample is not a multiple of `default_overall_up_factor`
        # forward_upsample_size = False
        # interpolate through upsample_size
        forward_upsample_size = True
        upsample_size = None

        for dim in sample.shape[-2:]:
            if dim % default_overall_up_factor != 0:
                # Forward upsample size to force interpolation output size.
                forward_upsample_size = True
                break

        # ensure attention_mask is a bias, and give it a singleton query_tokens dimension
        # expects mask of shape:
        #   [batch, key_tokens]
        # adds singleton query_tokens dimension:
        #   [batch,                    1, key_tokens]
        # this helps to broadcast it as a bias over attention scores, which will be in one of the following shapes:
        #   [batch,  heads, query_tokens, key_tokens] (e.g. torch sdp attn)
        #   [batch * heads, query_tokens, key_tokens] (e.g. xformers or classic attn)
        if attention_mask is not None:
            # assume that mask is expressed as:
            #   (1 = keep,      0 = discard)
            # convert mask into a bias that can be added to attention scores:
            #       (keep = +0,     discard = -10000.0)
            attention_mask = (1 - attention_mask.to(sample.dtype)) * -10000.0
            attention_mask = attention_mask.unsqueeze(1)

        # convert encoder_attention_mask to a bias the same way we do for attention_mask
        if encoder_attention_mask is not None:
            encoder_attention_mask = (
                1 - encoder_attention_mask.to(sample.dtype)
            ) * -10000.0
            encoder_attention_mask = encoder_attention_mask.unsqueeze(1)

        # 0. center input if necessary
        if self.config.center_input_sample:
            sample = 2 * sample - 1.0

        # 1. time
        timesteps = timestep
        if not torch.is_tensor(timesteps):
            # TODO: this requires sync between CPU and GPU. So try to pass timesteps as tensors if you can
            # This would be a good case for the `match` statement (Python 3.10+)
            is_mps = sample.device.type == "mps"
            if isinstance(timestep, float):
                dtype = torch.float32 if is_mps else torch.float64
            else:
                dtype = torch.int32 if is_mps else torch.int64
            timesteps = torch.tensor([timesteps], dtype=dtype, device=sample.device)
        elif len(timesteps.shape) == 0:
            timesteps = timesteps[None].to(sample.device)

        # broadcast to batch dimension in a way that's compatible with ONNX/Core ML
        # modified to support dynamic shape for onediff
        # timesteps = timesteps.expand(sample.shape[0])
        timesteps = torch._C.broadcast_dim_like(timesteps, sample, dim=0)

        t_emb = self.time_proj(timesteps)

        # `Timesteps` does not contain any weights and will always return f32 tensors
        # but time_embedding might actually be running in fp16. so we need to cast here.
        # there might be better ways to encapsulate this.
        t_emb = t_emb.to(dtype=sample.dtype)

        emb = self.time_embedding(t_emb, timestep_cond)
        aug_emb = None

        if self.class_embedding is not None:
            if class_labels is None:
                raise ValueError(
                    "class_labels should be provided when num_class_embeds > 0"
                )

            if self.config.class_embed_type == "timestep":
                class_labels = self.time_proj(class_labels)

                # `Timesteps` does not contain any weights and will always return f32 tensors
                # there might be better ways to encapsulate this.
                class_labels = class_labels.to(dtype=sample.dtype)

            class_emb = self.class_embedding(class_labels).to(dtype=sample.dtype)

            if self.config.class_embeddings_concat:
                emb = torch.cat([emb, class_emb], dim=-1)
            else:
                emb = emb + class_emb

        if self.config.addition_embed_type == "text":
            aug_emb = self.add_embedding(encoder_hidden_states)
        elif self.config.addition_embed_type == "text_image":
            # Kandinsky 2.1 - style
            if "image_embeds" not in added_cond_kwargs:
                raise ValueError(
                    f"{self.__class__} has the config param `addition_embed_type` set to 'text_image' which requires the keyword argument `image_embeds` to be passed in `added_cond_kwargs`"
                )

            image_embs = added_cond_kwargs.get("image_embeds")
            text_embs = added_cond_kwargs.get("text_embeds", encoder_hidden_states)
            aug_emb = self.add_embedding(text_embs, image_embs)
        elif self.config.addition_embed_type == "text_time":
            # SDXL - style
            if "text_embeds" not in added_cond_kwargs:
                raise ValueError(
                    f"{self.__class__} has the config param `addition_embed_type` set to 'text_time' which requires the keyword argument `text_embeds` to be passed in `added_cond_kwargs`"
                )
            text_embeds = added_cond_kwargs.get("text_embeds")
            if "time_ids" not in added_cond_kwargs:
                raise ValueError(
                    f"{self.__class__} has the config param `addition_embed_type` set to 'text_time' which requires the keyword argument `time_ids` to be passed in `added_cond_kwargs`"
                )
            time_ids = added_cond_kwargs.get("time_ids")
            time_embeds = self.add_time_proj(time_ids.flatten())
            # modified to support dynamic shape for onediff
            # time_embeds = time_embeds.reshape((text_embeds.shape[0], -1))
            time_embeds = time_embeds.unflatten(
                dim=0, shape=(-1, time_ids.shape[1])
            ).flatten(1, 2)
            add_embeds = torch.concat([text_embeds, time_embeds], dim=-1)
            add_embeds = add_embeds.to(emb.dtype)
            aug_emb = self.add_embedding(add_embeds)
        elif self.config.addition_embed_type == "image":
            # Kandinsky 2.2 - style
            if "image_embeds" not in added_cond_kwargs:
                raise ValueError(
                    f"{self.__class__} has the config param `addition_embed_type` set to 'image' which requires the keyword argument `image_embeds` to be passed in `added_cond_kwargs`"
                )
            image_embs = added_cond_kwargs.get("image_embeds")
            aug_emb = self.add_embedding(image_embs)
        elif self.config.addition_embed_type == "image_hint":
            # Kandinsky 2.2 - style
            if (
                "image_embeds" not in added_cond_kwargs
                or "hint" not in added_cond_kwargs
            ):
                raise ValueError(
                    f"{self.__class__} has the config param `addition_embed_type` set to 'image_hint' which requires the keyword arguments `image_embeds` and `hint` to be passed in `added_cond_kwargs`"
                )
            image_embs = added_cond_kwargs.get("image_embeds")
            hint = added_cond_kwargs.get("hint")
            aug_emb, hint = self.add_embedding(image_embs, hint)
            sample = torch.cat([sample, hint], dim=1)

        emb = emb + aug_emb if aug_emb is not None else emb

        if self.time_embed_act is not None:
            emb = self.time_embed_act(emb)

        if (
            self.encoder_hid_proj is not None
            and self.config.encoder_hid_dim_type == "text_proj"
        ):
            encoder_hidden_states = self.encoder_hid_proj(encoder_hidden_states)
        elif (
            self.encoder_hid_proj is not None
            and self.config.encoder_hid_dim_type == "text_image_proj"
        ):
            # Kadinsky 2.1 - style
            if "image_embeds" not in added_cond_kwargs:
                raise ValueError(
                    f"{self.__class__} has the config param `encoder_hid_dim_type` set to 'text_image_proj' which requires the keyword argument `image_embeds` to be passed in  `added_conditions`"
                )

            image_embeds = added_cond_kwargs.get("image_embeds")
            encoder_hidden_states = self.encoder_hid_proj(
                encoder_hidden_states, image_embeds
            )
        elif (
            self.encoder_hid_proj is not None
            and self.config.encoder_hid_dim_type == "image_proj"
        ):
            # Kandinsky 2.2 - style
            if "image_embeds" not in added_cond_kwargs:
                raise ValueError(
                    f"{self.__class__} has the config param `encoder_hid_dim_type` set to 'image_proj' which requires the keyword argument `image_embeds` to be passed in  `added_conditions`"
                )
            image_embeds = added_cond_kwargs.get("image_embeds")
            encoder_hidden_states = self.encoder_hid_proj(image_embeds)
        elif (
            self.encoder_hid_proj is not None
            and self.config.encoder_hid_dim_type == "ip_image_proj"
        ):
            if "image_embeds" not in added_cond_kwargs:
                raise ValueError(
                    f"{self.__class__} has the config param `encoder_hid_dim_type` set to 'ip_image_proj' which requires the keyword argument `image_embeds` to be passed in  `added_conditions`"
                )
            image_embeds = added_cond_kwargs.get("image_embeds")
            image_embeds = self.encoder_hid_proj(image_embeds).to(
                encoder_hidden_states.dtype
            )
            encoder_hidden_states = torch.cat(
                [encoder_hidden_states, image_embeds], dim=1
            )

        # 2. pre-process
        sample = self.conv_in(sample)

        # 2.5 GLIGEN position net
        if (
            cross_attention_kwargs is not None
            and cross_attention_kwargs.get("gligen", None) is not None
        ):
            cross_attention_kwargs = cross_attention_kwargs.copy()
            gligen_args = cross_attention_kwargs.pop("gligen")
            cross_attention_kwargs["gligen"] = {
                "objs": self.position_net(**gligen_args)
            }

        # 3. down
        lora_scale = (
            cross_attention_kwargs.get("scale", 1.0)
            if cross_attention_kwargs is not None
            else 1.0
        )
        if USE_PEFT_BACKEND:
            # weight the lora layers by setting `lora_scale` for each PEFT layer
            scale_lora_layers(self, lora_scale)

        is_controlnet = (
            mid_block_additional_residual is not None
            and down_block_additional_residuals is not None
        )
        # using new arg down_intrablock_additional_residuals for T2I-Adapters, to distinguish from controlnets
        is_adapter = down_intrablock_additional_residuals is not None
        # maintain backward compatibility for legacy usage, where
        #       T2I-Adapter and ControlNet both use down_block_additional_residuals arg
        #       but can only use one or the other
        if (
            not is_adapter
            and mid_block_additional_residual is None
            and down_block_additional_residuals is not None
        ):
            deprecate(
                "T2I should not use down_block_additional_residuals",
                "1.3.0",
                "Passing intrablock residual connections with `down_block_additional_residuals` is deprecated \
                       and will be removed in diffusers 1.3.0.  `down_block_additional_residuals` should only be used \
                       for ControlNet. Please make sure use `down_intrablock_additional_residuals` instead. ",
                standard_warn=False,
            )
            down_intrablock_additional_residuals = down_block_additional_residuals
            is_adapter = True

        down_block_res_samples = (sample,)
        for downsample_block in self.down_blocks:
            if (
                hasattr(downsample_block, "has_cross_attention")
                and downsample_block.has_cross_attention
            ):
                # For t2i-adapter CrossAttnDownBlock2D
                additional_residuals = {}
                if is_adapter and len(down_intrablock_additional_residuals) > 0:
                    additional_residuals[
                        "additional_residuals"
                    ] = down_intrablock_additional_residuals.pop(0)

                sample, res_samples = downsample_block(
                    hidden_states=sample,
                    temb=emb,
                    encoder_hidden_states=encoder_hidden_states,
                    attention_mask=attention_mask,
                    cross_attention_kwargs=cross_attention_kwargs,
                    encoder_attention_mask=encoder_attention_mask,
                    **additional_residuals,
                )
            else:
                if diffusers_version < diffusers_0210_v:
                    sample, res_samples = downsample_block(
                        hidden_states=sample, temb=emb
                    )
                else:
                    sample, res_samples = downsample_block(
                        hidden_states=sample, temb=emb, scale=lora_scale
                    )
                if is_adapter and len(down_intrablock_additional_residuals) > 0:
                    sample += down_intrablock_additional_residuals.pop(0)

            down_block_res_samples += res_samples

        if is_controlnet:
            new_down_block_res_samples = ()

            for down_block_res_sample, down_block_additional_residual in zip(
                down_block_res_samples, down_block_additional_residuals
            ):
                down_block_res_sample = (
                    down_block_res_sample + down_block_additional_residual
                )
                new_down_block_res_samples = new_down_block_res_samples + (
                    down_block_res_sample,
                )

            down_block_res_samples = new_down_block_res_samples

        # 4. mid
        if self.mid_block is not None:
            if (
                hasattr(self.mid_block, "has_cross_attention")
                and self.mid_block.has_cross_attention
            ):
                sample = self.mid_block(
                    sample,
                    emb,
                    encoder_hidden_states=encoder_hidden_states,
                    attention_mask=attention_mask,
                    cross_attention_kwargs=cross_attention_kwargs,
                    encoder_attention_mask=encoder_attention_mask,
                )
            else:
                sample = self.mid_block(sample, emb)

            # To support T2I-Adapter-XL
            if (
                is_adapter
                and len(down_intrablock_additional_residuals) > 0
                and sample.shape == down_intrablock_additional_residuals[0].shape
            ):
                sample += down_intrablock_additional_residuals.pop(0)

        if is_controlnet:
            sample = sample + mid_block_additional_residual

        # 5. up
        for i, upsample_block in enumerate(self.up_blocks):
            is_final_block = i == len(self.up_blocks) - 1

            res_samples = down_block_res_samples[-len(upsample_block.resnets) :]
            down_block_res_samples = down_block_res_samples[
                : -len(upsample_block.resnets)
            ]

            # if we have not reached the final block and need to forward the
            # upsample size, we do it here
            if not is_final_block and forward_upsample_size:
                # To support dynamic switching of special resolutions, pass a like tensor.
                upsample_size = down_block_res_samples[-1].shape[2:]

            output_like = None
            if not is_final_block:
                output_like = down_block_res_samples[-1]

            if (
                hasattr(upsample_block, "has_cross_attention")
                and upsample_block.has_cross_attention
            ):
                sample = upsample_block(
                    hidden_states=sample,
                    temb=emb,
                    res_hidden_states_tuple=res_samples,
                    encoder_hidden_states=encoder_hidden_states,
                    cross_attention_kwargs=cross_attention_kwargs,
                    upsample_size=upsample_size,
                    output_like=output_like,
                    attention_mask=attention_mask,
                    encoder_attention_mask=encoder_attention_mask,
                )
            else:
                if diffusers_version < diffusers_0210_v:
                    sample = upsample_block(
                        hidden_states=sample,
                        temb=emb,
                        res_hidden_states_tuple=res_samples,
                        upsample_size=upsample_size,
                        output_like=output_like,
                    )
                else:
                    sample = upsample_block(
                        hidden_states=sample,
                        temb=emb,
                        res_hidden_states_tuple=res_samples,
                        upsample_size=upsample_size,
                        output_like=output_like,
                        scale=lora_scale,
                    )

        # 6. post-process
        if self.conv_norm_out:
            sample = self.conv_norm_out(sample)
            sample = self.conv_act(sample)
        sample = self.conv_out(sample)

        if USE_PEFT_BACKEND:
            # remove `lora_scale` from each PEFT layer
            unscale_lora_layers(self, lora_scale)

        if not return_dict:
            return (sample,)

        return UNet2DConditionOutput(sample=sample)



from typing import Callable

import torch

from ..registry import register_backend
from .deployable_module import get_deployable_module, NexfortDeployableModule


@register_backend("nexfort")
def compile(torch_module: torch.nn.Module, *, options=None):
    from nexfort.compilers import nexfort_compile

    # Decorator mode
    if torch_module is None:

        def fn(torch_module: Callable):
            if torch_module is None:
                raise RuntimeError("torch_module can't be None")
            return compile(torch_module, options=options)

        return fn

    if isinstance(torch_module, NexfortDeployableModule):
        return compile(torch_module._torch_module, options=options)

    if isinstance(options, str):
        import json

        # TODO(): using jsonschema to define the options schema
        options = json.loads(options)

    nexfort_options = options if options is not None else dict()

    compiled_model = nexfort_compile(torch_module, **nexfort_options)

    return get_deployable_module(torch_module, compiled_model)



from typing import Any

import torch


class DeployableModule(torch.nn.Module):
    def __init__(self):
        torch.nn.Module.__init__(self)

    def forward(self, *args, **kwargs) -> Any:
        raise NotImplementedError()



import unittest

import numpy as np
from PIL import Image
from skimage.metrics import structural_similarity as ssim


class QuantizeQuality(unittest.TestCase):
    def test_validate(self):
        image1 = np.array(
            Image.open("/share_nfs/civitai/20240407-163408.jpg").convert("RGB")
        )
        image2 = np.array(
            Image.open("/src/onediff/output_enterprise_sd.png").convert("RGB")
        )
        # Calculate SSIM
        ssim_index = ssim(image1, image2, multichannel=True, win_size=3)
        print("SSIM:", ssim_index)
        self.assertTrue(
            ssim_index > 0.89, "SSIM Validation fails, and the workflow is aborted"
        )


if __name__ == "__main__":
    unittest.main()



import copy
import time
import unittest
from functools import partial

import onediff.infer_compiler as infer_compiler

import torch
from onediff.utils.import_utils import is_nexfort_available, is_oneflow_available


class SubModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = torch.nn.Linear(100, 10)

    def forward(self, x):
        return self.linear(x)


class MainModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = torch.nn.Linear(1000, 100)
        self.sub_module = SubModule()

    def forward(self, x):
        x = torch.nn.functional.relu(self.linear(x))
        x = self.sub_module(x)
        return x


def compute(x):
    return torch.sin(x) + torch.cos(x)


class TestModelInference(unittest.TestCase):
    def setUp(self) -> None:
        self.compilation_functions = []

        if is_oneflow_available():
            oneflow_compile_fn = partial(infer_compiler.compile, backend="oneflow")
            self.compilation_functions.append(oneflow_compile_fn)

        if is_nexfort_available():
            nexfort_compile_options = {
                "mode": "max-optimize:max-autotune:freezing:benchmark:cudagraphs",
                "dynamic": True,
                "fullgraph": True,
            }
            nexfort_compile_fn = partial(
                infer_compiler.compile,
                backend="nexfort",
                options=nexfort_compile_options,
            )
            self.compilation_functions.append(nexfort_compile_fn)

        assert len(self.compilation_functions) > 0

    def measure_inference_time(
        self, model, warmup=3, num_runs=30, input_args=[], input_kwargs={}
    ):
        for _ in range(warmup):
            model(*input_args, **input_kwargs)

        total_time = 0.0
        for _ in range(num_runs):
            start_time = time.time()
            model(*input_args, **input_kwargs)
            total_time += time.time() - start_time

        average_time = total_time / num_runs
        result = model(*input_args, **input_kwargs)
        return result, average_time

    def generate_models_and_inputs(self):
        for compile_fn in self.compilation_functions:
            model = MainModule().cuda().half()
            inputs = [torch.randn(10000, 1000).cuda().half()]
            compiled_model = compile_fn(model)
            yield model, compiled_model, inputs, {}

            model = MainModule().cuda().half()
            inputs = [torch.randn(10000, 1000).cuda().half()]
            compiled_model_sub = copy.deepcopy(model)
            compiled_model_sub.sub_module = compile_fn(compiled_model_sub.sub_module)
            yield model, compiled_model_sub, inputs, {}

            if compile_fn.keywords.get("backend") == "nexfort":
                inputs_compute = [torch.randn(10000, 1000).cuda().half()]
                compiled_compute_fn = compile_fn(compute)
                yield compute, compiled_compute_fn, inputs_compute, {}

    @torch.inference_mode()
    def test_inference_results(self):
        for (
            model,
            compiled_model,
            input_args,
            input_kwargs,
        ) in self.generate_models_and_inputs():
            original_result, _ = self.measure_inference_time(
                model, input_args=input_args, input_kwargs=input_kwargs
            )
            compiled_result, _ = self.measure_inference_time(
                compiled_model, input_args=input_args, input_kwargs=input_kwargs
            )

            self.assertTrue(
                torch.allclose(original_result, compiled_result, atol=1e-2, rtol=1e-3)
            )

            if isinstance(model, torch.nn.Module):
                self.assertIsInstance(compiled_model, MainModule)
                self.assertEqual(
                    set(model.state_dict().keys()),
                    set(compiled_model.state_dict().keys()),
                )


if __name__ == "__main__":
    unittest.main()



import os
import oneflow as flow  # usort: skip
import oneflow.nn as nn
import oneflow.nn.functional as F


class FusedSelfAttnProcessor:
    def __init__(self, attn):
        assert hasattr(attn, "to_qkv") and attn.to_qkv is not None

    def __call__(
        self,
        attn,
        hidden_states,
        encoder_hidden_states=None,
        attention_mask=None,
        temb=None,
    ):
        assert (
            encoder_hidden_states is None
        ), "encoder_hidden_states must be None for FusedSelfAttnProcessor"

        residual = hidden_states

        if attn.spatial_norm is not None:
            hidden_states = attn.spatial_norm(hidden_states, temb)

        input_ndim = hidden_states.ndim

        if input_ndim == 4:
            batch_size, channel, height, width = hidden_states.shape
            hidden_states = hidden_states.view(
                batch_size, channel, height * width
            ).transpose(1, 2)

        batch_size, sequence_length, _ = (
            hidden_states.shape
            if encoder_hidden_states is None
            else encoder_hidden_states.shape
        )

        if attention_mask is not None:
            attention_mask = attn.prepare_attention_mask(
                attention_mask, sequence_length, batch_size
            )
            # scaled_dot_product_attention expects attention_mask shape to be
            # (batch, heads, source_length, target_length)
            attention_mask = attention_mask.view(
                batch_size, attn.heads, -1, attention_mask.shape[-1]
            )

        if attn.group_norm is not None:
            hidden_states = attn.group_norm(hidden_states.transpose(1, 2)).transpose(
                1, 2
            )

        qkv = attn.to_qkv(hidden_states)

        inner_dim = qkv.shape[-1] // 3
        head_dim = inner_dim // attn.heads

        if False:
            qkv = qkv.view(-1, attn.heads, head_dim * 3)

            query = (
                qkv[:, :, 0:head_dim].reshape(batch_size, -1, inner_dim).contiguous()
            )
            key = (
                qkv[:, :, head_dim : head_dim * 2]
                .reshape(batch_size, -1, inner_dim)
                .contiguous()
            )
            value = (
                qkv[:, :, head_dim * 2 : head_dim * 3]
                .reshape(batch_size, -1, inner_dim)
                .contiguous()
            )

            query = attn.head_to_batch_dim(query)
            key = attn.head_to_batch_dim(key)
            value = attn.head_to_batch_dim(value)

            attention_probs = attn.get_attention_scores(query, key, attention_mask)
            hidden_states = flow.bmm(attention_probs, value)
            hidden_states = attn.batch_to_head_dim(hidden_states)
        else:
            from onediff.utils import parse_boolean_from_env, set_boolean_env_var

            if attn.upcast_attention and parse_boolean_from_env(
                "ONEFLOW_ATTENTION_ALLOW_HALF_PRECISION_ACCUMULATION", True
            ):
                set_boolean_env_var(
                    "ONEFLOW_ATTENTION_ALLOW_HALF_PRECISION_ACCUMULATION", False
                )
            hidden_states = flow._C.fused_multi_head_attention_inference_v2(
                query=qkv,
                query_head_size=head_dim,
                query_layout="BM(H3K)",
                output_layout="BM(HK)",
                scale=attn.scale,
                causal=False,
            )

        # linear proj
        hidden_states = attn.to_out[0](hidden_states)
        # dropout
        hidden_states = attn.to_out[1](hidden_states)

        if input_ndim == 4:
            hidden_states = hidden_states.transpose(-1, -2).reshape(
                batch_size, channel, height, width
            )

        if attn.residual_connection:
            hidden_states = hidden_states + residual

        hidden_states = hidden_states / attn.rescale_output_factor

        return hidden_states


try:
    from onediff.infer_compiler.backends.oneflow.transform import register

    def convert_fused_self_attn_processor(
        mod: FusedSelfAttnProcessor, verbose=True
    ) -> FusedSelfAttnProcessor:
        return mod

    register(torch2oflow_funcs=convert_fused_self_attn_processor)
except:
    print("Skip onediff.infer_compiler.backends.oneflow.transform.register")



import warnings
from collections import defaultdict
from typing import List, Union

import torch

from onediff.infer_compiler import DeployableModule

_nested_counter = defaultdict(lambda: 0)


class TensorInplaceAssign:
    r"""
    This class is used as a context manager, instantiated with either a `torch.nn.Module` or
    `onediff.infer_compiler.backends.deployable_module.DeployableModule` during initialization.
    Within the context manager, all Tensors associated with the provided module will be
    transformed into AutoInplaceCopyTensor. After transformed, assignments to Tensor.data are
    modified to in-place copying.

    The class is commonly used to ensure the stability of the data_ptr() for weights,
    particularly in scenarios involving the loading of LoRA weights.

    For example:
        >>> class EagerModule(torch.nn.Module):
        >>>     def __init__(self):
        >>>         super().__init__()
        >>>         self.linear1 = torch.nn.Linear(3, 3)
        >>>         self.linear2 = torch.nn.Linear(3, 3)
        >>>
        >>>     def forward(self, x):
        >>>         return self.linear2(self.linear1(x))
        >>>
        >>> eager = EagerModule()
        >>> dptr1 = eager.linear1.weight.data.data_ptr()
        >>> dptr2 = eager.linear2.weight.data.data_ptr()
        >>>
        >>> with TensorInplaceAssign(eager.linear1):
        >>>     eager.linear1.weight.data = torch.randn(3, 3)
        >>>     eager.linear2.weight.data = torch.randn(3, 3)
        >>>
        >>> eager.linear1.weight.data.data_ptr() == dptr1, eager.linear2.weight.data.data_ptr() == dptr2
        (True, False)
    """

    def __init__(
        self, *modules: List[Union[torch.nn.Module, DeployableModule]]
    ) -> None:
        warnings.warn(
            "The class TensorInplaceAssign is deprecated and will be removed soon. \
                      If you are using _fuse_lora in TensorInplaceAssign, please check `onediff.utils.load_and_fuse_lora`"
        )
        self.modules = set()
        for module in modules:
            if isinstance(module, torch.nn.Module):
                self.modules.add(module)
            elif isinstance(module, DeployableModule):
                self.modules.add(module._deployable_module_model._torch_module)
            else:
                raise TypeError(
                    "TensorInplaceAssign can only accept torch.nn.Module or DeployableModule"
                )

    def __enter__(self):
        global _nested_counter
        for module in self.modules:
            if _nested_counter[module] == 0:
                module.apply(module_convert_parameter)
            _nested_counter[module] += 1

    def __exit__(self, exc_type, exc_value, traceback):
        global _nested_counter
        for module in list(self.modules):
            _nested_counter[module] -= 1
            if _nested_counter[module] == 0:
                module.apply(module_unconvert_parameter)
                _nested_counter.pop(module)
                self.modules.remove(module)


class AutoInplaceCopyTensor(torch.Tensor):
    @property
    def data(self):
        return AutoInplaceCopyTensor(self)

    @data.setter
    def data(self, new_tensor):
        if not isinstance(new_tensor, torch.Tensor):
            raise TypeError(
                f"Cannot assign type {type(new_tensor)} to AutoInplaceCopyTensor"
            )
        self.copy_(new_tensor.detach())


class AutoInplaceCopyParameter(torch.nn.Parameter):
    @property
    def data(self):
        return AutoInplaceCopyTensor(super(AutoInplaceCopyParameter, self).data)

    @data.setter
    def data(self, new_tensor):
        if not isinstance(new_tensor, torch.Tensor):
            raise TypeError(
                f"Cannot assign type {type(new_tensor)} to AutoInplaceCopyParameter"
            )
        self.data.copy_(new_tensor.detach())


def module_convert_parameter(module: torch.nn.Module) -> torch.nn.Module:
    for k, v in module.__dict__.items():
        if isinstance(v, torch.nn.Parameter):
            module.__dict__[k] = AutoInplaceCopyParameter(v)
        elif isinstance(v, torch.Tensor):
            module.__dict__[k] = AutoInplaceCopyTensor(v)
    for k, param in module._parameters.items():
        if not isinstance(param, AutoInplaceCopyParameter) and param is not None:
            module._parameters[k] = AutoInplaceCopyParameter(param)
    for k, buffer in module._buffers.items():
        if not isinstance(buffer, AutoInplaceCopyTensor) and buffer is not None:
            module._buffers[k] = AutoInplaceCopyTensor(buffer)
    return module


def module_unconvert_parameter(module: torch.nn.Module) -> torch.nn.Module:
    for k, v in module.__dict__.items():
        if isinstance(v, AutoInplaceCopyParameter):
            module.__dict__[k] = torch.nn.Parameter(torch.Tensor(v.data))
        elif isinstance(v, AutoInplaceCopyTensor):
            module.__dict__[k] = torch.Tensor(v)
    for k, param in module._parameters.items():
        if isinstance(param, AutoInplaceCopyParameter):
            module._parameters[k] = torch.nn.Parameter(torch.Tensor(param.data))
    for k, buffer in module._buffers.items():
        if isinstance(buffer, AutoInplaceCopyTensor):
            module._buffers[k] = torch.Tensor(buffer)
    return module


if __name__ == "__main__":

    class EagerModule(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.linear1 = torch.nn.Linear(3, 3)
            self.linear2 = torch.nn.Linear(3, 3)

        def forward(self, x):
            return self.linear2(self.linear1(x))

    eager = EagerModule()
    dptr1 = eager.linear1.weight.data.data_ptr()
    dptr2 = eager.linear2.weight.data.data_ptr()

    with TensorInplaceAssign(eager):
        eager.linear1.weight.data = torch.randn(3, 3)
        eager.linear2.weight.data = torch.randn(3, 3)

    assert dptr1 == eager.linear1.weight.data.data_ptr()
    assert dptr2 == eager.linear2.weight.data.data_ptr()

    dptr1 = eager.linear1.weight.data.data_ptr()
    dptr2 = eager.linear2.weight.data.data_ptr()
    with TensorInplaceAssign(eager.linear1):
        eager.linear1.weight.data = torch.randn(3, 3)
        eager.linear2.weight.data = torch.randn(3, 3)
    assert dptr1 == eager.linear1.weight.data.data_ptr()
    assert dptr2 != eager.linear2.weight.data.data_ptr()

    dptr1 = eager.linear1.weight.data.data_ptr()
    dptr2 = eager.linear2.weight.data.data_ptr()
    with TensorInplaceAssign(eager.linear1):
        with TensorInplaceAssign(eager.linear2):
            with TensorInplaceAssign(eager.linear1):
                pass
            eager.linear1.weight.data = torch.randn(3, 3)
            eager.linear2.weight.data = torch.randn(3, 3)
    assert dptr1 == eager.linear1.weight.data.data_ptr()
    assert dptr2 == eager.linear2.weight.data.data_ptr()



_ONEFLOW_EXEC_MODE = False


class oneflow_exec_mode(object):
    def __init__(self, enabled=None):
        if enabled is not None:
            self.enabled = enabled
        else:
            self.enabled = True

    def __enter__(self):
        import oneflow as flow  # usort: skip

        global _ONEFLOW_EXEC_MODE
        self.prev_mode = _ONEFLOW_EXEC_MODE
        _ONEFLOW_EXEC_MODE = self.enabled
        self.prev_grad_mode = flow.is_grad_enabled()
        _ = flow.set_grad_enabled(False)

    def __exit__(self, exc_type, exc_val, exc_tb):
        import oneflow as flow  # usort: skip

        global _ONEFLOW_EXEC_MODE
        _ONEFLOW_EXEC_MODE = self.prev_mode
        _ = flow.set_grad_enabled(self.prev_grad_mode)


def oneflow_exec_mode_enabled():
    global _ONEFLOW_EXEC_MODE
    return _ONEFLOW_EXEC_MODE



import importlib
import os
import platform
import traceback
from inspect import ismodule
from types import ModuleType

system = platform.system()


def check_module_availability(module_name):
    spec = importlib.util.find_spec(module_name)

    if spec:
        try:
            importlib.import_module(module_name)
        except ImportError as e:
            print(traceback.format_exc())
            return False
    else:
        return False

    return True


_oneflow_available = check_module_availability("oneflow")
_onediff_quant_available = check_module_availability("onediff_quant")
_nexfort_available = check_module_availability("nexfort")

if system != "Linux":
    print(f"Warning: OneFlow is only supported on Linux. Current system: {system}")
    _oneflow_available = False


def is_oneflow_available():
    return _oneflow_available


def is_onediff_quant_available():
    return _onediff_quant_available


def is_nexfort_available():
    return _nexfort_available


class DynamicModuleLoader(ModuleType):
    def __init__(self, obj_entity: ModuleType, pkg_root=None, module_path=None):
        self.obj_entity = obj_entity
        self.pkg_root = pkg_root
        self.module_path = module_path

    @classmethod
    def from_path(cls, module_path: str):
        model_name = os.path.basename(module_path)
        module = importlib.import_module(model_name)
        return cls(module, module_path, module_path)

    def __getattr__(self, name):
        obj_entity = getattr(self.obj_entity, name, None)
        module_path = os.path.join(self.module_path, name)
        if obj_entity is None:
            pkg_name = os.path.basename(self.pkg_root)
            absolute_name = os.path.relpath(module_path, self.pkg_root).replace(
                os.path.sep, "."
            )
            absolute_name = f"{pkg_name}.{absolute_name}"
            obj_entity = importlib.import_module(absolute_name)
        if ismodule(obj_entity):
            return DynamicModuleLoader(obj_entity, self.pkg_root, module_path)
        return obj_entity



"""Module to convert PyTorch code to OneFlow."""
from .builtin_transform import (
    default_converter,
    get_attr,
    map_args,
    proxy_class,
    ProxySubmodule,
    torch2oflow,
)
from .custom_transform import register
from .manager import transform_mgr



import importlib
import os
from typing import Dict

import torch
import oneflow as flow  # usort: skip
from functools import wraps
from pathlib import Path

from oneflow.framework.args_tree import ArgsTree

from onediff.utils import logger
from ..env_var import OneflowCompileOptions
from .transform.builtin_transform import torch2oflow
from .transform.manager import transform_mgr
from .utils.cost_util import cost_time
from .utils.hash_utils import generate_input_structure_key, generate_model_structure_key


def _prepare_file_path(file_path):
    if isinstance(file_path, Path):
        file_path = str(file_path)
    if file_path.endswith(".graph"):
        file_path = file_path[:-6]
    return file_path


@cost_time(debug=transform_mgr.debug_mode, message="generate graph file name")
def generate_graph_file_name(file_path, deployable_module, args, kwargs):
    file_path = _prepare_file_path(file_path)
    args_tree = ArgsTree((args, kwargs), gen_name=False, tensor_type=torch.Tensor)
    input_structure_key = generate_input_structure_key(args_tree)
    model_structure_key = generate_model_structure_key(deployable_module)
    # Combine cache keys
    cache_key = f"{input_structure_key}_{model_structure_key}"
    return f"{file_path}_{cache_key}.graph"


def graph_file_management(func):
    @wraps(func)
    def wrapper(self: "OneflowDeployableModule", *args, **kwargs):
        compile_options = (
            self._deployable_module_options
            if hasattr(self, "_deployable_module_options")
            else OneflowCompileOptions()
        )
        graph_file = compile_options.graph_file
        is_first_load = self._load_graph_first_run and graph_file is not None

        if self._deployable_module_input_structure_key is None:
            args_tree = ArgsTree(
                (args, kwargs), gen_name=False, tensor_type=torch.Tensor
            )
            self._deployable_module_input_structure_key = generate_input_structure_key(
                args_tree
            )

        if is_first_load:
            self._load_graph_first_run = False
            input_structure_key = self._deployable_module_input_structure_key
            model_structure_key = generate_model_structure_key(deployable_module=self)
            file_path = _prepare_file_path(graph_file)
            graph_file = (
                f"{file_path}_{input_structure_key}_{model_structure_key}.graph"
            )

        def process_state_dict_before_saving(state_dict: Dict):
            nonlocal self, args, kwargs, graph_file
            graph = self.get_graph()
            if importlib.util.find_spec("register_comfy"):
                from register_comfy import CrossAttntionStateDictPatch as state_patch

                state_dict = state_patch.process_state_dict_before_saving(
                    state_dict, graph=graph
                )
            return state_dict

        def handle_graph_loading():
            nonlocal graph_file, compile_options, is_first_load
            if not is_first_load:
                return

            if not os.path.exists(graph_file):
                logger.info(
                    f"Graph file {graph_file} does not exist! Generating graph."
                )
            else:
                graph_device = compile_options.graph_file_device
                state_dict = flow.load(graph_file)
                self.load_graph(
                    graph_file, torch2oflow(graph_device), state_dict=state_dict
                )
                logger.info(f"Loaded graph file: {graph_file}")
                is_first_load = False

        def handle_graph_saving():
            nonlocal graph_file, compile_options, is_first_load
            if not is_first_load:
                return

            parent_dir = os.path.dirname(graph_file)
            if parent_dir != "":
                os.makedirs(parent_dir, exist_ok=True)

            # Avoid graph file conflicts
            if os.path.exists(graph_file):
                raise FileExistsError(f"File {graph_file} exists!")
            try:

                self.save_graph(
                    graph_file, process_state_dict=process_state_dict_before_saving
                )
                logger.info(f"Saved graph file: {graph_file}")

            except Exception as e:
                logger.error(f"Failed to save graph file: {graph_file}! {e}")

        if self._deployable_module_options.use_graph and is_first_load:
            handle_graph_loading()
            ret = func(self, *args, **kwargs)
            handle_graph_saving()
        else:
            ret = func(self, *args, **kwargs)

        return ret

    return wrapper



import importlib.metadata
from typing import Optional

import oneflow as torch
import oneflow.nn as nn
import oneflow.nn.functional as F

from onediff.infer_compiler.backends.oneflow.transform import transform_mgr
from packaging import version

transformed_diffusers = transform_mgr.transform_package("diffusers")

diffusers_0210_v = version.parse("0.21.0")
diffusers_0220_v = version.parse("0.22.0")
diffusers_version = version.parse(importlib.metadata.version("diffusers"))

if diffusers_version < diffusers_0210_v:

    class Upsample2D(transformed_diffusers.models.resnet.Upsample2D):
        def forward(
            self,
            hidden_states: torch.FloatTensor,
            output_size: Optional[int] = None,
            output_like: Optional[torch.FloatTensor] = None,
        ) -> torch.FloatTensor:
            assert hidden_states.shape[1] == self.channels

            if self.use_conv_transpose:
                return self.conv(hidden_states)

            # Cast to float32 to as 'upsample_nearest2d_out_frame' op does not support bfloat16
            # TODO(Suraj): Remove this cast once the issue is fixed in PyTorch
            # https://github.com/pytorch/pytorch/issues/86679
            dtype = hidden_states.dtype
            if dtype == torch.bfloat16:
                hidden_states = hidden_states.to(torch.float32)

            # upsample_nearest_nhwc fails with large batch sizes. see https://github.com/huggingface/diffusers/issues/984
            if hidden_states.shape[0] >= 64:
                hidden_states = hidden_states.contiguous()

            # if `output_size` is passed we force the interpolation output
            # size and do not make use of `scale_factor=2`
            if output_like is not None:
                hidden_states = F.interpolate_like(
                    hidden_states, like=output_like, mode="nearest"
                )
            elif output_size is None:
                hidden_states = F.interpolate(
                    hidden_states, scale_factor=2.0, mode="nearest"
                )
            else:
                hidden_states = F.interpolate(
                    hidden_states, size=output_size, mode="nearest"
                )

            # If the input is bfloat16, we cast back to bfloat16
            if dtype == torch.bfloat16:
                hidden_states = hidden_states.to(dtype)

            # TODO(Suraj, Patrick) - clean up after weight dicts are correctly renamed
            if self.use_conv:
                if self.name == "conv":
                    hidden_states = self.conv(hidden_states)
                else:
                    hidden_states = self.Conv2d_0(hidden_states)

            return hidden_states

else:

    class Upsample2D(transformed_diffusers.models.resnet.Upsample2D):
        def forward(
            self,
            hidden_states: torch.FloatTensor,
            output_size: Optional[int] = None,
            output_like: Optional[torch.FloatTensor] = None,
            scale: float = 1.0,
        ) -> torch.FloatTensor:
            assert hidden_states.shape[1] == self.channels

            if self.use_conv_transpose:
                return self.conv(hidden_states)

            # Cast to float32 to as 'upsample_nearest2d_out_frame' op does not support bfloat16
            # TODO(Suraj): Remove this cast once the issue is fixed in PyTorch
            # https://github.com/pytorch/pytorch/issues/86679
            dtype = hidden_states.dtype
            if dtype == torch.bfloat16:
                hidden_states = hidden_states.to(torch.float32)

            # upsample_nearest_nhwc fails with large batch sizes. see https://github.com/huggingface/diffusers/issues/984
            if hidden_states.shape[0] >= 64:
                hidden_states = hidden_states.contiguous()

            # if `output_size` is passed we force the interpolation output
            # size and do not make use of `scale_factor=2`
            if output_like is not None:
                hidden_states = F.interpolate_like(
                    hidden_states, like=output_like, mode="nearest"
                )
            elif output_size is None:
                hidden_states = F.interpolate(
                    hidden_states, scale_factor=2.0, mode="nearest"
                )
            else:
                hidden_states = F.interpolate(
                    hidden_states, size=output_size, mode="nearest"
                )

            # If the input is bfloat16, we cast back to bfloat16
            if dtype == torch.bfloat16:
                hidden_states = hidden_states.to(dtype)

            # TODO(Suraj, Patrick) - clean up after weight dicts are correctly renamed
            if self.use_conv:
                if self.name == "conv":
                    if diffusers_version < diffusers_0220_v:
                        if isinstance(
                            self.conv,
                            transformed_diffusers.models.lora.LoRACompatibleConv,
                        ):
                            hidden_states = self.conv(hidden_states, scale)
                        else:
                            hidden_states = self.conv(hidden_states)
                    else:
                        if (
                            isinstance(
                                self.conv,
                                transformed_diffusers.models.lora.LoRACompatibleConv,
                            )
                            and not transformed_diffusers.utils.USE_PEFT_BACKEND
                        ):
                            hidden_states = self.conv(hidden_states, scale)
                        else:
                            hidden_states = self.conv(hidden_states)
                else:
                    if diffusers_version < diffusers_0220_v:
                        if isinstance(
                            self.Conv2d_0,
                            transformed_diffusers.models.lora.LoRACompatibleConv,
                        ):
                            hidden_states = self.Conv2d_0(hidden_states, scale)
                        else:
                            hidden_states = self.Conv2d_0(hidden_states)
                    else:
                        if (
                            isinstance(
                                self.Conv2d_0,
                                transformed_diffusers.models.lora.LoRACompatibleConv,
                            )
                            and not transformed_diffusers.utils.USE_PEFT_BACKEND
                        ):
                            hidden_states = self.Conv2d_0(hidden_states, scale)
                        else:
                            hidden_states = self.Conv2d_0(hidden_states)

            return hidden_states



import logging
import os
import time
from pathlib import Path


class ColorFormatter(logging.Formatter):
    COLORS = {
        "DEBUG": "\033[34m",  # Blue
        "INFO": "\033[92m",  # green
        "WARNING": "\033[93m",  # Yellow
        "ERROR": "\033[91m",  # Red
        "CRITICAL": "\033[91m",  # Red
    }

    def format(self, record):
        log_message = super().format(record)
        color = self.COLORS.get(record.levelname, "\033[0m")  # Default to Reset color
        return f"{color}{log_message}\033[0m"


class ConfigurableLogger:
    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)

    def __getattr__(self, name):
        return getattr(self.logger, name)

    def configure_logging(self, name, level, log_dir=None, file_name=None):
        logger = logging.getLogger(name)

        if logger.hasHandlers() and len(logger.handlers) > 0:
            logger.warning("Logging handlers already exist for %s", name)
            return

        logger.setLevel(level)
        # Prevent messages from being propagated to the root logger
        logger.propagate = False

        if logger.hasHandlers():
            logger.warning("Logging handlers already exist for %s", name)
        else:
            # Create a console formatter and add it to a console handler
            console_formatter = ColorFormatter(
                fmt="%(levelname)s [%(asctime)s] %(pathname)s:%(lineno)d - %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )

            console_handler = logging.StreamHandler()
            console_handler.setFormatter(console_formatter)
            logger.addHandler(console_handler)

            # Create a file formatter and add it to a file handler if log_dir is provided
            if log_dir:
                log_dir = Path(log_dir)
                os.makedirs(log_dir, exist_ok=True)

                file_prefix = "{}_".format(
                    time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime())
                )

                if file_name:
                    log_file_name = file_prefix + file_name
                else:
                    log_file_name = file_prefix + name + ".log"

                log_file = log_dir / log_file_name
                file_formatter = logging.Formatter(
                    fmt="%(levelname)s [%(asctime)s] - %(message)s",
                    datefmt="%Y-%m-%d %H:%M:%S",
                )
                file_handler = logging.FileHandler(log_file, encoding="utf-8")
                file_handler.setFormatter(file_formatter)
                logger.addHandler(file_handler)

        self.logger = logger


logger = ConfigurableLogger()



"""A module for registering custom torch2oflow functions and classes."""
import importlib.util
import inspect
import sys
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

from onediff.utils import logger
from ..import_tools import import_module_from_path
from .builtin_transform import torch2oflow
from .manager import transform_mgr

__all__ = ["register"]


def register_torch2oflow_class(cls: type, replacement: type, verbose=True):
    try:
        key = transform_mgr.get_transformed_entity_name(cls)
        transform_mgr.update_class_proxies({key: replacement}, verbose=verbose)

    except Exception as e:
        logger.warning(f"Cannot register {cls=} {replacement=}. {e=}")


def register_torch2oflow_func(func, first_param_type=None, verbose=False):
    if first_param_type is None:
        params = inspect.signature(func).parameters
        first_param_type = params[list(params.keys())[0]].annotation
        if first_param_type == inspect._empty:
            logger.warning(f"Cannot register {func=} {first_param_type=}.")
    try:
        torch2oflow.register(first_param_type)(func)
        logger.debug(f"Register {func=} {first_param_type=}")
        if verbose:
            logger.info(f"Register {func=} {first_param_type=}")
    except Exception as e:
        logger.warning(f"Cannot register {func=} {first_param_type=}. {e=}")


def set_default_registry():
    def import_module_safely(module_path, module_name):

        if module_name in transform_mgr.loaded_modules:
            return
        try:
            import_module_from_path(module_path)
        except Exception as e:
            logger.warning(f"Failed to import {module_name} from {module_path}. {e=}")
        finally:
            transform_mgr.loaded_modules.add(module_name)

    # compiler_registry_path
    registry_path = Path(__file__).parents[5] / "infer_compiler_registry"

    if sys.modules.get("diffusers") is not None:
        if importlib.util.find_spec("diffusers") is not None:
            import_module_safely(
                registry_path / "register_diffusers", "register_diffusers"
            )

        if importlib.util.find_spec("diffusers_enterprise_lite"):
            import_module_safely(
                registry_path / "register_diffusers_enterprise_lite",
                "register_diffusers_enterprise_lite",
            )

    if importlib.util.find_spec("onediff_quant") is not None:
        import_module_safely(
            registry_path / "register_onediff_quant", "register_onediff_quant"
        )


def ensure_list(obj):
    if isinstance(obj, list):
        return obj
    return [obj]


def register(
    *,
    package_names: Optional[List[Union[Path, str]]] = None,
    torch2oflow_class_map: Optional[Dict[type, type]] = None,
    torch2oflow_funcs: Optional[List[Callable]] = None,
):
    if package_names:
        package_names = ensure_list(package_names)
        transform_mgr.load_class_proxies_from_packages(package_names)

    if torch2oflow_class_map:
        for torch_cls, of_cls in torch2oflow_class_map.items():
            register_torch2oflow_class(torch_cls, of_cls)

    if torch2oflow_funcs:
        torch2oflow_funcs = ensure_list(torch2oflow_funcs)
        for func in torch2oflow_funcs:
            register_torch2oflow_func(func)



import re
import types

import torch
import oneflow as flow  # usort: skip
from typing import Any, Dict, List, Union

from onediff.utils import logger


def parse_device(args: List[Any], kwargs: Dict[str, Any]):
    if "device" in kwargs:
        return kwargs["device"]
    for x in args:
        if isinstance(x, (flow.device, torch.device)):
            return x
        if x in ["cpu", "cuda"]:
            return x
    return None


def check_device(current_device, target_device) -> bool:
    def _convert(device):
        assert isinstance(device, (str, torch.device, flow.device))
        if isinstance(device, torch.device):
            index = device.index if device.index is not None else 0
            return flow.device(device.type, index)
        if isinstance(device, str):
            return flow.device(device)
        return device

    return _convert(current_device) == _convert(target_device)


# hooks and helper functions for constant folding conv weights

STATE_UPDATED_ATTR = "_onediff_state_updated"
CONSTANT_FOLDING_INFO_ATTR = "_onediff_constant_folding_info"
GRAPH_RELATED_TENSOR_ATTR = "_onediff_graph_related_tensor"


def init_state_update_attr(module: torch.nn.Module):
    from onediff.infer_compiler import DeployableModule

    if isinstance(module, DeployableModule):
        module = module._torch_module
    if not isinstance(module, torch.nn.Module):
        raise TypeError(f"module must be a torch.nn.Module, got {type(module)}")
    setattr(module, STATE_UPDATED_ATTR, False)


def set_constant_folded_conv_attr(
    deployable_module, constant_folding_info: Dict[str, flow.Tensor] = None
) -> None:
    from onediff.infer_compiler import DeployableModule

    if not isinstance(deployable_module, DeployableModule):
        raise TypeError(
            f"deployable_model must be a DeployableModule, got {type(deployable_module)}"
        )

    constant_folding_info = constant_folding_info or get_constant_folding_info(
        deployable_module
    )
    if constant_folding_info is None:
        return

    torch_module: torch.nn.Module = deployable_module._torch_module
    for submodule in torch_module.modules():
        if isinstance(submodule, torch.nn.Conv2d) and hasattr(
            submodule, GRAPH_RELATED_TENSOR_ATTR
        ):
            delattr(submodule, GRAPH_RELATED_TENSOR_ATTR)

    for weight_name, weight_tensor in constant_folding_info.items():
        submodule = deployable_module._torch_module.get_submodule(
            removesuffix(weight_name, ".weight")
        )
        object.__setattr__(submodule, GRAPH_RELATED_TENSOR_ATTR, weight_tensor)


def generate_constant_folding_info(
    deployable_module, torch_module: torch.nn.Module = None
) -> Dict[str, flow.Tensor]:
    removeprefix = lambda ss, prefix: ss[len(prefix) :] if ss.startswith(prefix) else ss

    # convert str like 'variable_transpose_model.input_blocks.10.0.in_layers.2.weight_239'
    # to 'input_blocks.10.0.in_layers.2.weight'
    def convert_var_name(s: str, prefix="variable_transpose_"):
        s = removeprefix(s, prefix)
        s = re.sub(r"_[0-9]+$", "", s)
        s = removeprefix(s, "model.")
        return s

    from onediff.infer_compiler import DeployableModule

    if not isinstance(deployable_module, DeployableModule):
        raise TypeError(
            f"deployable_model must be a DeployableModule, got {type(deployable_module)}"
        )
    if torch_module is None:
        torch_module = deployable_module._torch_module

    graph = deployable_module._deployable_module_dpl_graph
    if graph is None:
        raise RuntimeError(f"The graph of deployable_module is not built yet")

    result = {
        convert_var_name(k): v
        for k, v in zip(*graph._c_nn_graph.get_runtime_var_states())
        if k.startswith("variable_transpose_") and v.ndim == 4
    }

    setattr(deployable_module, CONSTANT_FOLDING_INFO_ATTR, result)

    set_constant_folded_conv_attr(deployable_module, result)

    def make_custom_copy_(module):
        def custom_copy_(self, src, non_blocking=False):
            torch.Tensor.copy_(self, src, non_blocking)
            # Update graph related tensors
            update_graph_related_tensor(module)

        return custom_copy_

    from onediff.torch_utils.module_operations import get_sub_module

    torch_model: torch.nn.Module = deployable_module._torch_module
    for k in result.keys():
        module = get_sub_module(torch_model, removesuffix(k, ".weight"))
        module.weight.copy_ = types.MethodType(make_custom_copy_(module), module.weight)


def update_graph_with_constant_folding_info(
    module: torch.nn.Module, info: Dict[str, flow.Tensor] = None
) -> None:
    from onediff.infer_compiler import DeployableModule

    if isinstance(module, DeployableModule):
        if info is None:
            info = get_constant_folding_info(module)
        module = module._torch_module
    if info is None:
        return

    for k in info:
        orig_tensor = module.get_parameter(k)
        target_tensor = info.get(k, None)
        if target_tensor is None:
            raise RuntimeError(f"Can't find tensor named {k} in graph")
        target_tensor.copy_(
            flow.utils.tensor.from_torch(orig_tensor.permute(0, 2, 3, 1))
        )


def update_graph_related_tensor(module: torch.nn.Conv2d) -> None:
    if not isinstance(module, torch.nn.Conv2d):
        return
    target_tensor = getattr(module, GRAPH_RELATED_TENSOR_ATTR, None)
    if target_tensor is None:
        return
    target_tensor.copy_(
        flow.utils.tensor.from_torch(module.weight.data.permute(0, 2, 3, 1))
    )


def get_constant_folding_info(module) -> Union[Dict[str, flow.Tensor], None]:
    from onediff.infer_compiler import DeployableModule

    if not isinstance(module, DeployableModule):
        raise TypeError(f"module must be a DeployableModule, got {type(module)}")
    return getattr(module, CONSTANT_FOLDING_INFO_ATTR, None)


def state_update_hook(module, incompatible_keys):
    if not hasattr(module, STATE_UPDATED_ATTR):
        return
    logger.info(f"load_state_dict called, set {STATE_UPDATED_ATTR} to True")
    setattr(module, STATE_UPDATED_ATTR, True)


def forward_generate_constant_folding_info_hook(module, args, output):
    if module._deployable_module_dpl_graph is None:
        return

    if getattr(module, CONSTANT_FOLDING_INFO_ATTR, None) is not None:
        return

    logger.info(f"generate constant folding info")
    generate_constant_folding_info(module)


def forward_pre_check_and_update_state_hook(module, args):
    if module._deployable_module_dpl_graph is None:
        return

    if not getattr(module._torch_module, STATE_UPDATED_ATTR, False):
        return

    constant_folding_info = getattr(module, CONSTANT_FOLDING_INFO_ATTR, None)
    if constant_folding_info is None:
        return

    logger.info(f"state_dict updated, modify the related weight in graph")
    update_graph_with_constant_folding_info(module, constant_folding_info)
    setattr(module._torch_module, STATE_UPDATED_ATTR, False)


def removesuffix(s: str, suffix: str) -> str:
    if s.endswith(suffix):
        return s[: len(s) - len(suffix)]
    else:
        return s



import os
from functools import partial
from typing import Any, List, Optional, Union

from onediff_quant import quantize_pipeline, save_quantized

from .quantize_utils import load_calibration_and_quantize_pipeline, setup_onediff_quant


class QuantPipeline:
    @classmethod
    def from_quantized(
        self,
        cls,
        quantized_model_name_or_path: Optional[Union[str, os.PathLike]],
        *args,
        **kwargs
    ):
        """load a quantized model.

        - Example:
          ```python
          from diffusers import AutoPipelineForText2Image
          pipe = QuantPipeline.from_quantized(
              AutoPipelineForText2Image, quantized_model_path, torch_dtype=torch.float16, variant="fp16", use_safetensors=True
          )

          pipe_kwargs = dict(
              prompt=args.prompt,
              height=1024,
              width=1024,
              num_inference_steps=30,
          )
          pipe(**pipe_kwargs)
          ```
        """
        setup_onediff_quant()
        pipe = cls.from_pretrained(quantized_model_name_or_path, *args, **kwargs)
        load_calibration_and_quantize_pipeline(
            os.path.join(str(quantized_model_name_or_path), "calibrate_info.txt"), pipe
        )
        return pipe

    @classmethod
    def from_pretrained(
        self,
        cls,
        pretrained_model_name_or_path: Optional[Union[str, os.PathLike]],
        *args,
        **kwargs
    ):
        """load a floating model that to be quantized as int8.

        - Example:
          ```python
          from diffusers import AutoPipelineForText2Image
          pipe = QuantPipeline.from_pretrained(
              AutoPipelineForText2Image, floatting_model_path, torch_dtype=torch.float16, variant="fp16", use_safetensors=True
          )
          pipe.to("cuda")

          pipe_kwargs = dict(
              prompt=args.prompt,
              height=1024,
              width=1024,
              num_inference_steps=30,
          )
          pipe.quantize(**pipe_kwargs,
              conv_compute_density_threshold=900,
              linear_compute_density_threshold=300,
              conv_ssim_threshold=0.985,
              linear_ssim_threshold=0.991,
              save_as_float=False,
              cache_dir=None)

          pipe.save_quantized(args.quantized_model, safe_serialization=True)
          ```
        """
        pipe = cls.from_pretrained(pretrained_model_name_or_path, *args, **kwargs)
        pipe.quantize = partial(quantize_pipeline, pipe)
        pipe.save_quantized = partial(save_quantized, pipe)
        return pipe

    @classmethod
    def from_single_file(
        self,
        cls,
        pretrained_model_name_or_path: Optional[Union[str, os.PathLike]],
        *args,
        **kwargs
    ):
        pipe = cls.from_single_file(pretrained_model_name_or_path, *args, **kwargs)
        pipe.quantize = partial(quantize_pipeline, pipe)
        pipe.save_quantized = partial(save_quantized, pipe)
        return pipe



import torch
import oneflow as flow  # usort: skip
from oneflow.framework.args_tree import ArgsTree

from onediff.utils import logger

from .utils.hash_utils import generate_input_structure_key


def input_output_processor(func):
    def process_input(*args, **kwargs):
        def input_fn(value):
            if isinstance(value, torch.Tensor):
                # TODO: https://github.com/siliconflow/sd-team/issues/109
                return flow.utils.tensor.from_torch(value.contiguous())
            else:
                return value

        args_tree = ArgsTree((args, kwargs), False, tensor_type=torch.Tensor)

        input_structure_key = generate_input_structure_key(args_tree)
        out = args_tree.map_leaf(input_fn)
        mapped_args = out[0]
        mapped_kwargs = out[1]
        return mapped_args, mapped_kwargs, input_structure_key

    def process_output(output):
        def output_fn(value):
            if isinstance(value, flow.Tensor):
                return flow.utils.tensor.to_torch(value)
            else:
                return value

        out_tree = ArgsTree((output, None), False)
        out = out_tree.map_leaf(output_fn)
        return out[0]

    def wrapper(self: "OneflowDeployableModule", *args, **kwargs):
        mapped_args, mapped_kwargs, input_structure_key = process_input(*args, **kwargs)
        if (
            self._deployable_module_options.use_graph
            and self._deployable_module_enable_dynamic
            and self._deployable_module_dpl_graph is not None
            and self._deployable_module_input_structure_key != input_structure_key
        ):
            # Retrieve the deployable module graph from cache using the input structure key
            dpl_graph = self._deployable_module_graph_cache.get(
                input_structure_key, None
            )
            self._deployable_module_graph_cache.put(
                self._deployable_module_input_structure_key,
                self._deployable_module_dpl_graph,
            )

            # If a cached graph is found, update the deployable module graph and input structure key
            if dpl_graph is not None:
                self._deployable_module_dpl_graph = dpl_graph
                self._deployable_module_input_structure_key = input_structure_key
            else:
                logger.warning(
                    f"Input structure key {self._deployable_module_input_structure_key} to {input_structure_key} has changed. Resetting the deployable module graph. This may slow down the process."
                )
                self._deployable_module_dpl_graph = None
                self._deployable_module_input_structure_key = None
                self._load_graph_first_run = True

        output = func(self, *mapped_args, **mapped_kwargs)
        return process_output(output)

    return wrapper



import functools
import importlib
import os
import types
from typing import Any, cast, Dict, List, Optional, Sequence, Tuple


_BACKENDS: Dict[str, Any] = dict()


def register_backend(
    name: Optional[str] = None,
    tags: Sequence[str] = (),
):
    def wrapper(compiler_fn: Optional[Any] = None):
        if compiler_fn is None:
            return functools.partial(register_backend, name=name, tags=tags)
        assert callable(compiler_fn)
        fname = name or compiler_fn.__name__
        assert fname not in _BACKENDS, f"duplicate name: {fname}"
        _BACKENDS[fname] = compiler_fn
        compiler_fn._tags = tuple(tags)
        return compiler_fn

    return wrapper


def lookup_backend(compiler_fn):
    """Expand backend strings to functions"""
    if isinstance(compiler_fn, str):
        if compiler_fn not in _BACKENDS:
            _lazy_import(compiler_fn)
        if compiler_fn not in _BACKENDS:
            raise RuntimeError(f"invalid backend {compiler_fn}")
        compiler_fn = _BACKENDS[compiler_fn]
    return compiler_fn


def _lazy_import(backend_name):
    from .. import backends

    backend_path = f"{backends.__name__}.{backend_name}"
    importlib.import_module(backend_path)



import argparse

import torch
from diffusers import AutoPipelineForText2Image

from onediff.infer_compiler import oneflow_compile
from onediff.quantization.quantize_pipeline import QuantPipeline


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--prompt", default="a photo of an astronaut riding a horse on mars"
    )
    parser.add_argument("--height", type=int, default=1024)
    parser.add_argument("--width", type=int, default=1024)
    parser.add_argument("--num_inference_steps", type=int, default=30)
    parser.add_argument("--quantized_model", type=str, required=True)
    return parser.parse_args()


args = parse_args()

pipe = QuantPipeline.from_quantized(
    AutoPipelineForText2Image,
    args.quantized_model,
    torch_dtype=torch.float16,
    variant="fp16",
    use_safetensors=True,
)
pipe = pipe.to("cuda")

pipe_kwargs = dict(
    prompt=args.prompt,
    height=args.height,
    width=args.width,
    num_inference_steps=args.num_inference_steps,
)

pipe.unet = oneflow_compile(pipe.unet)
pipe(**pipe_kwargs).images[0].save("test.png")



from types import FunctionType
from typing import Type, Union

import torch
from torch import nn

from ..deployable_module import DeployableModule


class NexfortDeployableModule(DeployableModule):
    def __init__(self, compiled_module, torch_module):
        torch.nn.Module.__init__(self)
        object.__setattr__(self, "_torch_module", torch_module)
        object.__setattr__(self, "_deployable_module_model", compiled_module)
        # https://github.com/pytorch/pytorch/blob/main/torch/_dynamo/eval_frame.py#L148
        if isinstance(torch_module, nn.Module) and isinstance(
            compiled_module, torch._dynamo.eval_frame.OptimizedModule
        ):
            object.__setattr__(self, "_modules", compiled_module._orig_mod._modules)
            object.__setattr__(
                self, "_parameters", compiled_module._orig_mod._parameters
            )
            object.__setattr__(self, "_buffers", compiled_module._orig_mod._buffers)

    def forward(self, *args, **kwargs):
        with torch._dynamo.utils.disable_cache_limit():
            return self._deployable_module_model(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self._deployable_module_model, name)


def _create_deployable_function(
    compiled_model, torch_module: FunctionType = None
) -> FunctionType:
    def deploy_function(*args, **kwargs):
        with torch._dynamo.utils.disable_cache_limit():
            return compiled_model(*args, **kwargs)

    return deploy_function


def _create_mixed_deployable_module(
    compiled_model, torch_module: nn.Module
) -> Type[NexfortDeployableModule]:
    module_cls = type(torch_module)

    class MixedNexfortDeployableModule(NexfortDeployableModule, module_cls):
        def __init__(self, compiled_module, torch_module):
            super().__init__(compiled_module, torch_module)

        def _get_name(self):
            return f"{self.__class__.__name__}(of {module_cls.__name__})"

    return MixedNexfortDeployableModule(
        compiled_module=compiled_model, torch_module=torch_module
    )


def get_deployable_module(
    torch_module: Union[nn.Module, FunctionType], compiled_model
) -> Union[Type[NexfortDeployableModule], FunctionType]:
    if not isinstance(torch_module, nn.Module):
        return _create_deployable_function(compiled_model, torch_module)
    return _create_mixed_deployable_module(compiled_model, torch_module)
